/************************************************************************

    nixieclock - a nixie desktop timepiece
    Copyright (C) 2000 Greg Ercolano, erco@3dsite.com

    This program is free software; you can redistribute it and/or modify
    it under the terms of the GNU General Public License as published by
    the Free Software Foundation; either version 2 of the License, or
    (at your option) any later version.

    This program is distributed in the hope that it will be useful,
    but WITHOUT ANY WARRANTY; without even the implied warranty of
    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
    GNU General Public License for more details.

    You should have received a copy of the GNU General Public License
    along with this program; if not, write to the Free Software
    Foundation, Inc., 675 Mass Ave, Cambridge, MA 02139, USA.

*************************************************************************/

#include <time.h>
#include <ctype.h>	// tolower()
#include <stdio.h>
#include <string.h>
#ifndef _WIN32
#include <unistd.h>
#include <stdlib.h>
#else /*_WIN32*/
#include <windows.h>
#include <FL/x.h>	// fl_xid
#define popen _popen
#define pclose _pclose
#endif /*_WIN32*/
#include <FL/Fl.H>
#include <FL/Fl_Double_Window.H>
#include <FL/Fl_Pixmap.H>
#include <FL/Fl_Box.H>
#include <FL/fl_draw.H>
#include <FL/fl_ask.H>

#include "NixieDisplay.H"

// GLOBALS
NixieDisplay *G_nixiedisplay = 0;		// main window
static char 
           G_test = 0,				// 1=run alphanumeric test
	   G_above = 0,				// 1=always above
	  *G_timestring = 0,			// str2ftime string
	  *G_command = 0,			// optional command to run
	   G_verbose = 0;			// verbose output
float      G_speed = 1.0;			// update every second
time_t	   G_countdown = 0;			// countdown time if non-zero

void HelpAndExit()
{
    Fl::error(
        "nixieclock %s - a clock program showing time with Nixie tube alphanumerics\n"
        "(C) Copyright Greg Ercolano, 2000. This software is GPL'ed.\n"
	"\n"
	"USAGE\n"
	"    nixieclock [options]\n"
	"\n"
	"OPTIONS\n"
	"    -d              # digit date mode (eg. 07:57 07-30-01)\n"
	"    -s <timestr>    # use strftime(3) date format\n"
	"    -t              # test pattern for alphanumeric display\n"
	"    -n              # no background mode (don't fork)\n"
	"    -geometry <pos> # sets geometry (x windows style positions)\n"
// NOT YET
//	"    -c <epochtime>  # counts down to epochtime\n"

// NOT YET
//	"    -r cmd [args]   # runs command, each line of output updates in clock\n"
//	"    -v              # enables verbose mode for -r\n"
	"    -a              # always above (keeps clock above all other windows)\n"
	"    -h              # help\n"
	"\n"
	"EXAMPLES\n"
	"    nixieclock                               -- 07:57 Sat Feb 2000\n"
	"    nixieclock -d                            -- 07:57 07-30-01\n"
#ifdef _WIN32
	"    nixieclock -s %%c                         -- Locale specific: 01/01/00 23:59:59\n"
#else
	"    nixieclock -s %%c                         -- Locale specific: Mon Jan 01 23:59:59 2000\n"
#endif
	"    nixieclock -s \"%%I:%%M %%p %%m/%%d/%%y\"        -- 12:59 PM 12/31/00\n"
	"    nixieclock -s \"%%H:%%M:%%S\"                 -- 23:59:59 (counts seconds)\n"
	"    nixieclock -s \"%%H:%%M:%%S %%a %%b %%d\"        -- 23:59:59 Sat Dec 31\n"
	"    nixieclock -s \"%%a %%b %%m %%H:%%M:%%S %%Y\"     -- Unix format; Mon Jan 01 23:59:59 2000\n"
	"    nixieclock -geometry -0-0                -- position lower right\n"
	"    nixieclock -geometry -5-5                -- position 5 pixels lower right\n"
	"    nixieclock -geometry +0-0                -- position lower left\n"
// NOT YET
//	"    nixieclock -r 'vmstat 3' -geom 1280x40   -- vmstat appears in clock\n"
	"\n", VERSION);

    exit(1);
}

void Tick_CB(void *data)
{
    // NEW TIME STRING
    char s[80];
    time_t timt = time(NULL);
    struct tm *tmbuf = localtime(&timt);
    strftime(s, sizeof(s)-1, G_timestring, tmbuf);
    G_nixiedisplay->SetString(s);

    // REPEAT TIMER
    Fl::repeat_timeout(G_speed, Tick_CB, data);
}

// FIGURE OUT DIGIT WIDTH OF SPECIFIED TIME STRING
int TimeStringWidth(const char *timestring)
{
    char s[80];
    time_t timt = time(NULL);
    struct tm *tmbuf = localtime(&timt);
    strftime(s, sizeof(s)-1, G_timestring, tmbuf);
    return(strlen(s));
}

// DISABLE BORDER AROUND WINDOW
void NoBorder()
{
    // DON'T CLEAR BORDER ON DARWIN UNLESS FLTK 1.1.7 OR HIGHER
    // eg:
    //     FL_MAJOR_VERSION        1
    //     FL_MINOR_VERSION        1
    //     FL_PATCH_VERSION        7
    //
#if !defined(DARWIN) || \
       ( FL_MAJOR_VERSION >= 1 && FL_MINOR_VERSION > 1 ) || \
       ( FL_MAJOR_VERSION == 1 && FL_MINOR_VERSION == 1 && FL_PATCH_VERSION >= 7 )

    G_nixiedisplay->clear_border();

#endif
}

// MAIN
int main(int argc, char **argv) 
{
    // BANNER
    fprintf(stderr, "\nnixieclock %s\n"
                    "(C) Copyright 2002 Greg Ercolano (erco@3dsite.com).\n"
		    "This code is GPL'ed.\n\n", VERSION);

    int user_geometry = 0;

    // PARSE ARGUMENTS
    {
	int nofork = 0, arg_offset = 0;
	for ( int t=1; t<argc; t++ )
	{
	    // HELP
	    if ( strcmp(argv[t], "-help") == 0 ||
	         strcmp(argv[t], "--help") == 0) HelpAndExit();

	    if ( strcmp(argv[t], "-geometry") == 0 )
	        { user_geometry = 1; continue; }

	    if ( argv[t][0] == '-' )
	    {
		if ( argv[t][0] == 0 ) 
		    HelpAndExit();

		switch ( argv[t][1] )
		{
		    case 'a': 
			G_above = 1;
			break;

		    case 't': 
			G_test = 1;
			break;

		    case 'n': 
			nofork = 1; 
			break;

		    case 's':
		    	if ( ++t >= argc ) 
			    { fprintf(stderr, "-s: missing argument\n"); exit(1); }
			G_timestring = argv[t];
			G_speed = 1.0;		// fast update
			++arg_offset;
			break;

		    // 23:59
		    case 'h': G_timestring="%H:%M"; break;

		    // 23:59 12/31/99
		    case 'd': G_timestring="%H:%M %m/%d/%y"; break;

		    // 23:59 Sat Dec 31
		    case 'm': G_timestring="%H:%M %a %b %d"; break; 

		    case 'c':
		    	if ( ++t >= argc )
			    { fprintf(stderr, "-c: missing argument\n"); exit(1); }
			G_countdown = strtol(argv[t], NULL, 0);
			++arg_offset;
			break;

		    case 'r':
		    	if ( ++t >= argc )
			    { fprintf(stderr, "-r: missing argument\n"); exit(1); }
		        G_command = argv[t];
			++arg_offset;
			break;

		    case 'v':
		        G_verbose = 1;
			break;

		    default: continue;
		}
		++arg_offset;
	    }
	}

	// PREVENT FLTK FROM SEEING ARGUMENTS WE PARSE
	argc -= arg_offset;
	argv += arg_offset;

#ifndef _WIN32
#ifndef DARWIN
	// UNIX: RUN IN THE BACKGROUND
	if ( ! nofork )
	{
	    switch ( fork() )
	    {
		case -1: perror("fork"); exit(1); 	// ERROR
		case  0: break;				// CHILD
		default: exit(0);			// PARENT DISCONNECTS
	    }
	}
#endif /*DARWIN*/
#endif /*_WIN32*/
    }

    // DEFAULT TIME USED IF UNSPECIFIED
    if ( G_timestring == NULL ) 
	G_timestring = "%H:%M %a %b %d";

    // FIGURE OUT DIGIT WIDTH NEEDED
    int numdigits = TimeStringWidth(G_timestring);

    // SET VISUAL MODE
    //
    //     FL_RGB prevents color map eating under X windows.
    //
    //     FL_DOUBLE|FL_INDEX enables hardware double buffering,
    //     but looks bad on Indy; digit backgrounds are gray.
    //     Fltk does a good job emulating double buffering
    //     in Fl_Double_Window w/out flickering, so leave off.
    //
    Fl::visual(FL_RGB);

    // PASS OTHER ARGS TO FLTK (-geometry, etc)
    Fl::args(argc, argv);

    // MAKE MAIN WINDOW
    int w = (FONTWIDTH + 2) * numdigits + ( XMARGIN * 2 );
    int h = FONTHEIGHT + ( YMARGIN * 2 );

    // DEFAULT DISPLAY TO LOWER RIGHT CORNER
    if ( ! user_geometry )
	G_nixiedisplay = new NixieDisplay(Fl::w()-w-5, Fl::h()-h-5, w, h,
					 "nixieclock");	// lower right
    else
	G_nixiedisplay = new NixieDisplay(w, h, "nixieclock");

    if ( G_speed > 1 )
	G_nixiedisplay->UpdateSpeed(1.0);	// slow update speed
    else
	G_nixiedisplay->UpdateSpeed(0.25);	// fast update

    NoBorder();

    //// BAD FOR WINDOWS (causes clock to appear in upperleft until "touched")
    //// G_nixiedisplay->free_position();

    G_nixiedisplay->show(argc,argv);

    // ALWAYS DRAW ABOVE
    if ( G_above )
    {
#ifdef _WIN32
         int x = SetWindowPos(fl_xid(G_nixiedisplay), HWND_TOPMOST, 0, 0,
                              0, 0, SWP_NOSIZE|SWP_NOMOVE);
#endif /*_WIN32*/
    }

    // SETUP TIMER TO UPDATE CLOCK'S TIME STRING
    Fl::add_timeout(G_speed, Tick_CB, (void*)G_nixiedisplay);

    return(Fl::run());
}
